;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; i386impl.m4 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;; i386 specific definitions for the LLL kernel ;;;;;;;;;;;;;;;;;;

include(as86.m4)
include(version)
include(magic.m4)

; See ~moose/info/i386impl.info for more info about this.

errprint({Definitions for the i386 generic kernel.
})dnl


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Structures ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Struc({AS},{
; Signature_Struc
  _long_(Magic_Number)		; magic number from magic file
  _byte_(Object_Name,20)
  _byte_(Version,8)
})

Struc({AH},{
; API_Header_Struc
; size is 32. offset is 32 from beginning of API file
; (32 first bytes are signature and version)
  _long_(TotalLength)		; total length of the file
  _long_(EntryPoint)		; address of object to return
  _long_(Offset)		; offset where to load it
  _long_(HeaderSize)		; size of header fields in file
  _long_(ExpandedSize)		; size when expanded to memory
  _long_()			; fill in
  _long_(CheckSum)		; checksum
  _long_(LongCheckSum)		; long checksum
})


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Some values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TRUE		=	-1	; Logical True value
FALSE		=	 0	; Logical False value

BYTEBITS	=	 8	; bits in a byte (smallest addressable value)

CHARABITS	=	 0	; bits of char in address
WORDABITS	=	 1	; bits of word in address
CELLABITS	=	 2	; bits of cell in address
PAGEABITS	=	12	; bits of page in address


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Deduced values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CHARSIZE	=	(1 << CHARABITS)	; size of a char in bytes
WORDSIZE	=	(1 << WORDABITS)	; size of a char in bytes
CELLSIZE	=	(1 << CELLABITS)	; size of a cell in bytes
PAGESIZE	=	(1 << PAGEABITS)	; size of a page in bytes
CHARMASK	=	~ (CHARSIZE - 1)	; mask of chars in address
WORDMASK	=	~ (WORDSIZE - 1)	; mask of words in address
CELLMASK	=	~ (CELLSIZE - 1)	; mask of cells in address
PAGEMASK	=	~ (PAGESIZE - 1)	; mask of pages in address
BITSPERCHAR	=	CHARSIZE*BYTEBITS	; bits in a char
BITSPERCELL	=	CELLSIZE*BYTEBITS	; bits in a cell


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;; abstraction of the i386 instruction set... ;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define({CPU},{i386})

; Registers:
; caller-saved: eax,ecx,edx
; callee-saved: ebx,esi,edi,ebp,esp
; constant:	cs,ds,es,ss,fs,gs
;  eax,edx	scratch; not saved upon a thread switch.
;  ecx		must be constant zero
;  ebx		buffers top of stack; saved upon a thread switch.
;  edi		data stack pointer (growing down)
;		If the stack is declared garbage-collection aware
;		all values on the stack may be corrected according to the
;		garbage collection on a "PAUSE", a "YIELD", a memory
;		allocation call or any such thing.
;		 And the thread will be unavailable during garbage
;		collection. Else, no relocatable address may be remembered
;		from a previous yield (), and none may be used at all if the
;		thread is preemptible.
;		Due to buffering, the stack must contain a ghost element
;		to be loaded into ebx when empty.
;  esp		local return stack pointer (growing down),
;		also heap pointer
;  ebp		frame on the heap
;  esi		points on current instruction to execute...
;
;
;
; Values:
; - de -1GB  +1GB-1
; - sinon, temporaire.
; pointeurs:
; - de +1GB  +3GB-1
; - sinon, non relogeable.


define({CELL},	CELLSIZE)		; a cell is 4 bytes long


;;; abstraction of the i386 instruction set

define({Terminator},__nl__)

define({Add},		{add $1,$2})
define({AddIm},		{lea $1,[$1+$2]})
define({Sub},		{sub $1,$2})
define({__Mov__},	{mov $1,$2})
define({__Movb__},	{movb $1,$2})
define({__Movw__},	{movw $1,$2})
define({Neg},		{neg $1})
define({Not},		{not $1})
define({And},		{and $1,$2})
define({Or},		{or $1,$2})
define({Xor},		{xor $1,$2})
define({Clr},		{clr($1)})
define({__Shr__},	{})

define({point},	{ifelse({$2},{},{[$1]},{[$1+CELL*$2]})})

define({PC},		{esi})	; program counter
define({DSP},		{edi})	; data stack pointer
define({DSINC},		{4})	; increment for data stack -- grow up
define({TOS},		{ebx})	; buffered top of stack
define({TOSL},		{bx})	; its lowest word
define({TOSLL},		{bl})	; its lowest byte
define({TOSLH},		{bh})	; its second byte
define({RSP},		{esp})	; return stack pointer
define({RSINC},		{(-4)})	; increment for return stack -- grow down

define({FRAME},		{ebp})	; a frame pointer for heap manipulation
define({R0},		{eax})	; a scratch register
define({R0L},		{ax})	; its lowest word
define({R0LL},		{al})	; its lowest byte
define({R0LH},		{ah})	; its second byte
define({R1},		{edx})	; second scratch register
define({COUNT},		{ecx})	; count register
define({COUNTLO},	{cl})	; count register -- low byte

define({INSTRTBL},	InstrT)		; table for bytecode threading

define({PopData},	{Mov({$1},TOS)
			 AddIm(DSP,-DSINC)})
define({PushData},	{AddIm(DSP,DSINC)
			 Mov(point(DSP),TOS)
			 Mov(TOS,{$1})})
define({PopReturn},	{ifelse({$1},,{AddIm(RSP,-RSINC)},{pop $1})})
define({PushReturn},	{push $1})
define({},		{})
define({},		{})
define({},		{})
define({},		{})

define({Enter},	{AddIm(PC,CELL)})
define({Next},	{jmp [PC]})


;;;;;;;;;;;;;;;;;;;;;;;;;; Defining FORTH words ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; defining a counted string
define({DefCString},
{__stringseg__({
locals({strlen},{strbeg},{strend},{
strbeg:
	.byte 	strlen
	defb($@)
strend:
strlen	=	strend-strbeg-1
})})})

;; defining a FORTH word


define({DefWord},{
ifelse({$2},,{$1},{$2}):
})

; flags for FORTH words:
;FW_IMMED	= 0x80	; word is immediate
;FW_INLINE	= 0x01	; word can be inlined


;;;;;;;;;;;;;;;;;;;;;;;;;;;; Absolute jump macros ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({jmpa},{
ifelse({$2},,
	{push #$1
	 ret},
	{mov $2,#$1
	 jmp $2}
)})
define({calla},{
ifelse({$2},,
	{mov eax,#$1
	 call eax},
	{mov $2,#$1
	 call $2}
)})


;;;;;;;;;;;;;;;;;;;;;;; Aerial (extra-light) threads ;;;;;;;;;;;;;;;;;;;;;;;;;
Struc({AT},{
; AerialThread	; These threads are lighter than light: Aerial !
  _long_(SAVE_EBX)	; initial Top of Stack
  _long_(SAVE_ESI)	; Data Stack Pointer
  _long_(SAVE_EDI)	; Return Stack Pointer
  _long_(SAVE_EBP)	; Those two are reserved for the
  _long_(SAVE_ESP)	;; threading mechanism.
  _long_(SAVE_RESP)	; ESP save when going ret-threaded
  _long_(STATE)		; flags used for thread sync'
  _long_(MAX_ESI)	; bottom of data stack
  _long_(MAX_EDI)	; bottom of return stack
  _long_(LINKBACK)	; Link Back to thread object
})


forget({
;------------------------- Signature --------------------------
;
; The first 32 bytes are reserved for the signature
; The first 20 bytes may be searched for to find the Kernel.
; The 12 remaining bytes may be used for version identification.
;
  .org 0
   MagicNumber		.long	0x01824B26FH
	; big-endian 	386	i	L	S	O
        ;       00  0110000010	01001	01100	10011	01111
				; 5 bit code for 'OSL' big-endian,
				; 5 bit code for i, 10-bit code for 386
				; two clear bits.
   Signature		.ascii	"OSLm Kernel i386 v0.00.01"
			.byte	0

;------------------------ Sizes -------------------------------
  ;; (!) The first 32 bytes were reserved for the signature.
  .org 32
			dd	TotalKernelFileLength
			dd	KernelOffset
			dd	KInitJumpPoint
			dd	size KHeader
})
